// AsmStats.exe
//
// Compiles statistics on types within a specified individual
// assembly, or all assemblies matching *.dll or *.exe
// in a given directory (optionally recursing through all
// subdirectories).
//
// This program isn't that useful in and of itself, but does
// demonstrates some simple reflection techniques.
//
// Refer to ShowCmdLineUsage for command line syntax.
//
// Mike Woodring
// http://staff.develop.com/woodring
//
using System;
using System.Reflection;
using System.Runtime.InteropServices;
using System.IO;

class AsmStatsApp
{
    static Type attributeClassType = typeof(Attribute);
    static Type delegateClassType = typeof(Delegate);

    #region AsmStats structure
    struct AsmStats
    {
        public string       FileSpec;            // a file path
        public ulong        NumAssemblies;
        public ulong        NumModules;
        public ulong        NumTypes;           // # of types of any kind
        public ulong        NumMembers;         // # of members of any kind
        public ulong        NumClasses;         // # of classes that are not delegates, interfaces, or attributes
        public ulong        NumPubClasses;      // # of classes that are public
        public ulong        NumAttributes;      // # of attribute classes
        public ulong        NumPubAttributes;   // # of those that are attributes
        public ulong        NumInterfaces;      // # of interfaces total
        public ulong        NumPubInterfaces;   // # of those that are public
        public ulong        NumValueTypes;      // # of value types total
        public ulong        NumPubValueTypes;   // # of those that are public
        public ulong        NumDelegates;
        public ulong        NumPubDelegates;
        public ulong        NumEnums;
        public ulong        NumPubEnums;
        public ulong        NumInstanceMethods;
        public ulong        NumStaticMethods;
        public ulong        NumInstanceProperties;
        public ulong        NumStaticProperties;
        public ulong        NumInstanceEvents;
        public ulong        NumStaticEvents;
        public ulong        NumInstanceFields;
        public ulong        NumStaticFields;
    }
    #endregion

    static void Main(string[] args)
    {
        (new AsmStatsApp())._Main(args);
    }

    void ShowCmdLineUsage()
    {
        Console.WriteLine("AsmStats command line usage.");
        Console.WriteLine("AsmStats FilePathOrDirectoryPathOrAsmName [/r(ecurse)]");
        Console.WriteLine("Examples:");
        Console.WriteLine("  asmstats foo.dll");
        Console.WriteLine("  asmstats mscorlib");
        Console.WriteLine("  asmstats \"System.Xml, Version=1.0.3300.0, Culture=neutral, PublicKeyToken=..\"");
        Console.WriteLine(@"  asmstats c:\winnt\system32");
        Console.WriteLine(@"  asmstats c:\winnt\system32 /r");
    }

    void _Main( string[] args )
    {
        if( args.Length < 1 )
        {
            ShowCmdLineUsage();
            return;
        }

        Console.WriteLine();
        Console.Out.Flush();
        DirectConsole.SetTitle("Assembly Statistics Viewer");

        string fileSpec = args[0];
        AsmStats stats = new AsmStats();
        stats.FileSpec = fileSpec;

        if( File.Exists(fileSpec) )
        {
            // It's a file; do it.
            //
            GetStats(fileSpec, ref stats);
            ShowStats(stats);
        }
        else if( Directory.Exists(fileSpec) )
        {
            // It's a directory; process *.dll and *.exe.
            //
            ProcessDirectory(ref stats, fileSpec, (args.Length == 2));
            ShowStats(stats);
        }
        else
        {
            // See if the command line argument can be treated
            // as an assembly name (for example: asmstats mscorlib).
            // If Assembly.Load can find it, extract the file path
            // to that assembly and then process that the normal
            // way.
            //
            string asmLocation = null;

            try
            {
                Assembly a = Assembly.Load(fileSpec);
                asmLocation = a.Location;
            }
            catch
            {
            }

            if( asmLocation != null )
            {
                GetStats(asmLocation, ref stats);
                ShowStats(stats);
            }
            else
            {
                Console.WriteLine("Argument is not a valid file or directory.");
            }
        }
    }

    void ProcessDirectory( ref AsmStats stats, string dirName, bool recurse )
    {
        string[] extensions = {"*.exe", "*.dll"};

        foreach( string ext in extensions )
        {
            string[] files = Directory.GetFiles(dirName, ext);

            foreach( string file in files )
            {
                GetStats(file, ref stats);
            }
        }

        if( recurse )
        {
            string[] subDirs = Directory.GetDirectories(dirName);

            foreach( string subDir in subDirs )
            {
                ProcessDirectory(ref stats, subDir, true);
            }
        }
    }

    void ShowStats( AsmStats stats )
    {
        Console.WriteLine();

        Console.WriteLine("Done processing {0}.", stats.FileSpec);
        Console.WriteLine( "Processed {0} {1} comprising {2} {3}.",
                           stats.NumAssemblies,
                           (stats.NumAssemblies == 1 ? "assembly" : "assemblies"),
                           stats.NumModules,
                           (stats.NumModules == 1 ? "module" : "modules") );
        Console.WriteLine();
        Console.WriteLine("  Types:         {0:n0} (of any kind)", stats.NumTypes);
        Console.WriteLine("    Classes:     {0:n0} ({1:n0} public)", stats.NumClasses, stats.NumPubClasses);
        Console.WriteLine("    Attributes:  {0:n0} ({1:n0} public)", stats.NumAttributes, stats.NumPubAttributes);
        Console.WriteLine("    Delegates:   {0:n0} ({1:n0} public)", stats.NumDelegates, stats.NumPubDelegates);
        Console.WriteLine("    Interfaces:  {0:n0} ({1:n0} public)", stats.NumInterfaces, stats.NumPubInterfaces);
        Console.WriteLine("    Enums:       {0:n0} ({1:n0} public)", stats.NumEnums, stats.NumPubEnums);
        Console.WriteLine("    Value types: {0:n0} ({1:n0} public)", stats.NumValueTypes, stats.NumPubValueTypes);
        Console.WriteLine();
        Console.WriteLine("  Members:       {0:n0} (of any kind, instance and static)", stats.NumMembers);
        Console.WriteLine("    Methods:     {0:n0} instance, {1:n0} static", stats.NumInstanceMethods, stats.NumStaticMethods);
        Console.WriteLine("    Events:      {0:n0} instance, {1:n0} static", stats.NumInstanceEvents, stats.NumStaticEvents);
        Console.WriteLine("    Properties:  {0:n0} instance, {1:n0} static", stats.NumInstanceProperties, stats.NumStaticProperties);
        Console.WriteLine("    Fields:      {0:n0} instance, {1:n0} static", stats.NumInstanceFields, stats.NumStaticFields);
    }

    string BaseFileName( string fileName )
    {
        int posLastSlash = fileName.LastIndexOf('\\');

        if( posLastSlash != -1 )
        {
            return fileName.Substring(posLastSlash + 1);
        }

        return(fileName);
    }

    void GetStats( string asmFileName, ref AsmStats stats )
    {
        StartNewAssemblyStatus(BaseFileName(asmFileName));

        try
        {
            Assembly a = Assembly.LoadFrom(asmFileName);
            stats.NumAssemblies++;
            stats.NumModules += (ulong)a.GetModules().Length;

            Type[] types = a.GetTypes();
            
            ulong totalTypes = (ulong)types.Length;
            ulong numTypesProcessed = 0;

            stats.NumTypes += (ulong)totalTypes;

            foreach( Type t in types )
            {
                if( t.IsInterface )
                {
                    stats.NumInterfaces++;

                    if( t.IsPublic )
                    {
                        stats.NumPubInterfaces++;
                    }
                }
                else if( t.IsEnum )
                {
                    stats.NumEnums++;

                    if( t.IsPublic )
                    {
                        stats.NumPubEnums++;
                    }
                }
                else if( t.IsValueType )
                {
                    stats.NumValueTypes++;

                    if( t.IsPublic )
                    {
                        stats.NumPubValueTypes++;
                    }
                }
                else if( t.IsClass )
                {
                    // Is this class an attribute, a delegate, or just
                    // a regular class?
                    //
                    if( attributeClassType.IsAssignableFrom(t) )
                    {
                        stats.NumAttributes++;

                        if( t.IsPublic )
                        {
                            stats.NumPubAttributes++;
                        }
                    }
                    else if( delegateClassType.IsAssignableFrom(t) )
                    {
                        stats.NumDelegates++;

                        if( t.IsPublic )
                        {
                            stats.NumPubDelegates++;
                        }
                    }
                    else
                    {
                        stats.NumClasses++;

                        if( t.IsPublic )
                        {
                            stats.NumPubClasses++;
                        }
                    }
                }
                else
                {
                    // System.Enum isn't categorized as a value type, enum,
                    // class, or interface according to the Type instance.
                    // So this app counts it as a value type.
                    //
                    if( t == typeof(System.Enum) )
                    {
                        stats.NumValueTypes++;
                        stats.NumPubValueTypes++;
                    }
                }

                MemberInfoShim memberShim = new MemberInfoShim(t);

                // Instance and static methods.
                //
                TallyMembers( new GetMembersProc(memberShim.GetMethods),
                              ref stats,
                              ref stats.NumInstanceMethods,
                              ref stats.NumStaticMethods );

                // Instance and static properties.
                //
                TallyMembers( new GetMembersProc(memberShim.GetProperties),
                              ref stats,
                              ref stats.NumInstanceProperties,
                              ref stats.NumStaticProperties );

                // Instance and static fields.
                //
                TallyMembers( new GetMembersProc(memberShim.GetFields),
                              ref stats,
                              ref stats.NumInstanceFields,
                              ref stats.NumStaticFields );

                // Instance and static events.
                //
                TallyMembers( new GetMembersProc(memberShim.GetEvents),
                              ref stats,
                              ref stats.NumInstanceEvents,
                              ref stats.NumStaticEvents );

                numTypesProcessed++;
                UpdateStatus((int)(((double)numTypesProcessed / (double)totalTypes) * 100));
            }
        }
        catch
        {
            UpdateStatus("could not process");
        }

        EndAssemblyStatus();
    }

    void TallyMembers( GetMembersProc GetMembers,
                       ref AsmStats stats,
                       ref ulong numInstanceMembers,
                       ref ulong numStaticMembers )
    {
        // Lookup and tally instance members.
        //
        BindingFlags bf = BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic;
        MemberInfo[] members = GetMembers(bf);
        ulong MemberCount = (ulong)(members != null ? members.Length : 0);
        stats.NumMembers += MemberCount;
        numInstanceMembers += MemberCount;

        // Lookup and tally static members.
        //
        bf = BindingFlags.Static | BindingFlags.Public | BindingFlags.NonPublic;
        members = GetMembers(bf);
        MemberCount = (ulong)(members != null ? members.Length : 0);
        stats.NumMembers += MemberCount;
        numStaticMembers += MemberCount;
    }

    // A little console state management for updating
    // this program's progress.
    //
    DirectConsole.COORD cpStatus;

    void StartNewAssemblyStatus( string asmName )
    {
        Console.Write("Processing {0}: ", asmName);
        cpStatus = DirectConsole.GetCursorPos();
    }

    void UpdateStatus( int percentComplete )
    {
        DirectConsole.SetCursorPosition(cpStatus);
        Console.Write("{0}%  ", percentComplete);
    }

    void UpdateStatus( string msg )
    {
        DirectConsole.SetCursorPosition(cpStatus);
        Console.Write(msg);
    }

    void EndAssemblyStatus()
    {
        Console.WriteLine();
    }
}

#region MemberInfoShim class

// The MemberInfoShim class and corresponding GetMembersProc
// delegate help me work around the fact that you cannot
// bind a delegate that looks like GetMembersProc to
// something like Type.GetFields, because the signature of
// GetFields (FieldInfo[](BindingFlasg)) isn't an exact
// match.  So even though the returned FieldInfo array *is*
// type compatible with the more generic MemberInfo array
// type, you cannot construct a delegate like GetMembersProc
// and bind it to GetFields/Methods/Properties/Events.
//
// Thanks for Valery Pryamikov for suggestion the following
// elegant work around.
//
delegate MemberInfo[] GetMembersProc( BindingFlags bf );

class MemberInfoShim
{
    public MemberInfoShim( Type t )
    {
        targetType = t;
    }

    public MemberInfo[] GetMethods( BindingFlags bf )
    {
        return targetType.GetMethods(bf);
    }

    public MemberInfo[] GetProperties( BindingFlags bf )
    {
        return targetType.GetProperties(bf);
    }

    public MemberInfo[] GetEvents( BindingFlags bf )
    {
        return targetType.GetEvents(bf);
    }

    public MemberInfo[] GetFields( BindingFlags bf )
    {
        return targetType.GetFields(bf);
    }

    Type targetType;
}
#endregion

#region DirectConsole class

// DirectConsole
//
// This class lets me play games with the console cursor
// position and window caption.
//
class DirectConsole
{
    [StructLayout(LayoutKind.Sequential, Pack=8)]
    public struct COORD
    {
        public COORD( short x, short y )
        {
            this.x = x;
            this.y = y;
        }

        public short x;
        public short y;
    }

    public enum StdHandle : ulong
    {
        STD_INPUT_HANDLE = unchecked((ulong)-10),
        STD_OUTPUT_HANDLE = unchecked((ulong)-11),
        STD_ERROR_HANDLE = unchecked((ulong)-12)
    }

    [StructLayout(LayoutKind.Sequential, Pack=8)]
    public struct SMALL_RECT
    {
        public short Left;
        public short Right;
        public short Top;
        public short Bottom;
    }

    [StructLayout(LayoutKind.Sequential, Pack=8)]
    public struct CONSOLE_SCREEN_BUFFER_INFO
    {
        public COORD        CursorSize;
        public COORD        CursorPosition;
        public ushort       Attributes;
        public SMALL_RECT   WindowRect;
        public COORD        MaxWindowSize;
    }

    [DllImport("kernel32.dll", EntryPoint="SetConsoleTitle", CharSet=CharSet.Auto)]
    public static extern bool SetTitle( string caption );

    [DllImport("kernel32.dll", EntryPoint="SetConsoleCursorPosition")]
    public static extern bool SetCursorPosition( int hConsole, COORD pos );

    public static bool SetCursorPosition( COORD pos )
    {
        return SetCursorPosition(GetStdOut(), pos);
    }

    [DllImport("kernel32.dll")]
    public static extern int GetStdHandle( StdHandle whichOne );

    public static int GetStdOut()
    {
        return GetStdHandle(StdHandle.STD_OUTPUT_HANDLE);
    }

    [DllImport("kernel32.dll", EntryPoint="GetConsoleScreenBufferInfo")]
    public static extern bool GetScreenBufferInfo(
                                  int hConsole,
                                  [Out] out CONSOLE_SCREEN_BUFFER_INFO info );

    public static COORD GetCursorPos( int hConsole )
    {
        CONSOLE_SCREEN_BUFFER_INFO csbi = new CONSOLE_SCREEN_BUFFER_INFO();
        GetScreenBufferInfo(hConsole, out csbi);
        return(csbi.CursorPosition);
    }

    public static COORD GetCursorPos()
    {
        return GetCursorPos(GetStdOut());
    }
}

#endregion // DirectConsole
